/* Extended Attribute support for rsync */
/* Copyright (C) 2004 Red Hat, Inc */
/* Written by Jay Fenlason, vaguely based on the ACLs patch */

/* This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "rsync.h"

#if SUPPORT_XATTRS

extern int preserve_xattrs;
extern int dry_run;

#define RSYNC_XAL_INITIAL 5
#define RSYNC_XAL_LIST_INITIAL 100


typedef struct {
	size_t name_len;
	char *name;
	size_t datum_len;
	char *datum;
} rsync_xa;

typedef struct {
	size_t count;
	size_t alloc;
	rsync_xa *rxas;
} rsync_xal;

typedef struct {
    size_t count;
    size_t alloc;
    rsync_xal *rxals;
} rsync_xal_list;

static size_t namebuf_len = 0;
static char *namebuf = NULL;

static size_t datumbuf_len = 0;
static char *datumbuf = NULL;

static rsync_xal curr_rsync_xal = { 0, 0, NULL };
static rsync_xal_list rsync_xal_l = { 0, 0, NULL };


/* ------------------------------------------------------------------------- */

/* the below stuff is only used by the receiver */

/* structure to hold index to rsync_xal_l member corresponding to
 * flist->files[i] */

typedef struct {
	const struct file_struct *file;
	int xalidx;
} file_xal_index;

typedef struct {
	size_t count;
	size_t alloc;
	file_xal_index *filexalidxs;
} file_xal_index_list;

static file_xal_index_list fxil = {0, 0, NULL };

/* stuff for redirecting calls to set_acl() from set_perms()
 * for keep_backup() */
static const struct file_struct *backup_orig_file = NULL;
static const char null_string[] = "";
static const char *backup_orig_fname = null_string;
static const char *backup_dest_fname = null_string;
static rsync_xal backup_xal;

/* ------------------------------------------------------------------------- */

static void rsync_xal_free ( rsync_xal *x ) {
	size_t i;

	for ( i = 0; i < x->count; i++ ) {
		free ( x->rxas[i].name );
		/* free ( x->rxas[i].value ); */
	}
	x->count = 0;
}

static int rsync_xal_compare_names ( const void *x1, const void *x2 ) {
	const rsync_xa *xa1;
	const rsync_xa *xa2;

	xa1 = x1;
	xa2 = x2;
	return strcmp ( xa1->name, xa2->name );
}

static BOOL rsync_xal_get ( const char *fname, rsync_xal *x ) {
	ssize_t name_size;
	ssize_t datum_size;
	ssize_t left;
	char *name;
	size_t len;
	char *ptr;

	if ( ! namebuf ) {
		namebuf_len = 100;
		namebuf = malloc ( 100 );
		datumbuf_len = 100;
		datumbuf = malloc ( 100 );
		if ( ! namebuf || ! datumbuf ) {
			rprintf(FERROR, "%s: rsync_xal_get: allocate intial buffers failed: %s\n",
 fname, strerror(errno));
			return False;
		}
	}

	name_size = sys_llistxattr ( fname, namebuf, namebuf_len );
	if ( name_size > namebuf_len ) {
	    name_size = -1;
	    errno = ERANGE;
	}
	if ( name_size < 0 ) {
	    if ( errno == ENOTSUP )
		return False;
	    if ( errno == ERANGE ) {
		name_size = sys_llistxattr ( fname, NULL, 0 );
		if ( name_size < 0 ) {
			rprintf ( FERROR, "%s: rsync_xal_get: llistxattr: %s\n",
 fname, strerror(errno) );
			return False;
		}
		namebuf = realloc ( namebuf, name_size + 1 );
		if ( ! namebuf ) {
		    rprintf ( FERROR, "%s: rsync_xal_get: reallocate namebuf to %lu failed: %s\n",
 fname, (unsigned long)name_size, strerror(errno) );
		    return False;
		}
		namebuf_len = name_size;
		name_size = sys_llistxattr ( fname, namebuf, namebuf_len );
		if ( name_size < 0 ) {
		    rprintf ( FERROR, "%s: rsync_xal_get: re-llistxattr failed: %s\n",
 fname, strerror(errno) );
		    return False;
		}
	    } else {
		rprintf(FERROR, "%s: rsync_xal_get: llistxattr failed: %s\n",
 fname, strerror(errno));
		return False;
	    }
	}
	rsync_xal_free ( x );
	if ( name_size == 0 )
	    return True;
	for ( left = name_size, name = namebuf; left > 0 ; left -= len, name += len ) {
		len = strlen ( name ) + 1;

		if ( x->count >= x->alloc ) {
			size_t new_alloc;
			size_t new_size;
			rsync_xa *new_rxas;

			new_alloc = ( x->alloc < RSYNC_XAL_INITIAL ) ? RSYNC_XAL_INITIAL : x->alloc * 2;
			new_size = new_alloc * sizeof(rsync_xa);
			new_rxas = x->alloc ? realloc ( x->rxas, new_size ) : malloc ( new_size );
			if ( new_rxas == NULL ) {
				rprintf(FERROR, "%s: rsync_xal_get: allocate %lufailed: %s\n",
 fname, (unsigned long)new_size, strerror(errno));
				return False;
			}
			x->alloc = new_alloc;
			x->rxas = new_rxas;
		}
		datum_size = sys_lgetxattr ( fname, name, datumbuf, datumbuf_len );
		if ( datum_size > datumbuf_len ) {
		    datum_size = -1;
		    errno = ERANGE;
		}
		if ( datum_size < 0 ) {
			if ( errno == ENOTSUP )
				return False;
			if ( errno == ERANGE ) {
				datum_size = sys_lgetxattr ( fname, name, NULL, 0 );
				if ( datum_size < 0 ) {
					rprintf ( FERROR, "%s: rsync_xal_get: lgetxattr %s failed: %s\n",
 fname, name, strerror(errno) );
					return False;
				}
				datumbuf = realloc ( datumbuf, datum_size + 1 );
				if ( ! datumbuf ) {
					rprintf(FERROR, "%s: rsync_xal_get: allocate %lu for datumbuf failed: %s\n",
 fname, (unsigned long)(datum_size + 1), strerror(errno));
					return False;
				}
				datumbuf_len = datum_size;
				datum_size = sys_lgetxattr ( fname, name, datumbuf, datumbuf_len );
				if ( datum_size < 0 ) {
					rprintf ( FERROR, "%s: rsync_xal_get: re-lgetxattr of %s failed: %s\n",
 name, fname, strerror(errno) );
					return False;
				}
			} else {
				rprintf(FERROR, "%s: rsync_xal_get: lgetxattr %s failed: %s\n",
 fname, name, strerror(errno));
				return False;
			}
		}
		ptr = malloc ( len + datum_size );
		strcpy ( ptr, name );
		if ( datum_size )
			memcpy ( ptr + len, datumbuf, datum_size );
		x->rxas[curr_rsync_xal.count].name_len = len;
		x->rxas[curr_rsync_xal.count].name = ptr;
		x->rxas[curr_rsync_xal.count].datum_len = datum_size;
		x->rxas[curr_rsync_xal.count].datum = ptr + len;
		x->count++;
	}
	if ( x->count > 1 ) {
		qsort ( x->rxas, x->count, sizeof(rsync_xa), rsync_xal_compare_names );
	}
	return True;
}


/* generate the xattr(s) for this flist entry;
 * xattr(s) are either sent or cleaned-up by send_xattr() below */

BOOL make_xattr(const struct file_struct *file, const char *fname)
{
	if (!preserve_xattrs)
		return True;

	rsync_xal_get ( fname, &curr_rsync_xal );
	return True;
}

static ssize_t rsync_xal_find_matching ( void ) {
    size_t i;
    size_t j;

    for ( i = 0; i < rsync_xal_l.count; i++ ) {
	/* Wrong number of elements? */
	if ( rsync_xal_l.rxals[i].count != curr_rsync_xal.count )
	    continue;
	/* any elements different? */
	for ( j = 0; j < curr_rsync_xal.count; j++ ) {
	    if (
 rsync_xal_l.rxals[i].rxas[j].name_len != curr_rsync_xal.rxas[j].name_len
 || rsync_xal_l.rxals[i].rxas[j].datum_len != curr_rsync_xal.rxas[j].datum_len
 || strcmp ( rsync_xal_l.rxals[i].rxas[j].name, curr_rsync_xal.rxas[j].name )
 || memcmp ( rsync_xal_l.rxals[i].rxas[j].datum, curr_rsync_xal.rxas[j].datum, curr_rsync_xal.rxas[j].datum_len ) )
		break;
	}
	/* no differences found.  This is The One! */
	if ( j == curr_rsync_xal.count )
	    break;
    }
    if ( i < rsync_xal_l.count )
	return i;
    return (ssize_t)-1;
}

/* Store curr_rsync_xal on the end of rsync_xal_l */
static void rsync_xal_store ( void ) {
	if ( rsync_xal_l.count <= rsync_xal_l.alloc ) {
		size_t new_alloc;
		size_t new_size;
		void *new_xal;

		new_alloc = ( rsync_xal_l.count < RSYNC_XAL_LIST_INITIAL ) ? RSYNC_XAL_LIST_INITIAL : rsync_xal_l.count * 2;
		new_size = new_alloc * sizeof(rsync_xal);
		new_xal = rsync_xal_l.count ? realloc ( rsync_xal_l.rxals, new_size ) : malloc ( new_size );
		if ( new_xal == NULL ) {
			rprintf ( FERROR, "rsync_xal_store: allocate %lu for new_xal failed\n", (unsigned long)new_size );
			exit_cleanup(RERR_STREAMIO );
		}
		rsync_xal_l.alloc = new_alloc;
		rsync_xal_l.rxals = new_xal;
	}
	rsync_xal_l.rxals[rsync_xal_l.count] = curr_rsync_xal;
	rsync_xal_l.count++;
	curr_rsync_xal.count = 0;
	curr_rsync_xal.alloc = 0;
}

/* send the make_xattr()-generated xattr list for this flist entry,
 * or clean up after an flist entry that's not being sent (f == -1) */

void send_xattr ( const struct file_struct *file, int f )
{
	ssize_t index;

	if ( ! preserve_xattrs )
		return;

	if (f == -1) {
		rsync_xal_free ( &curr_rsync_xal );
		return;
	}
	index = rsync_xal_find_matching ( );
	if ( index != -1) {
		write_byte(f, 'x' );
		write_int(f, index);
		rsync_xal_free ( &curr_rsync_xal );
	} else {
		rsync_xa *rxa;
		size_t count;

		count = curr_rsync_xal.count;
		write_byte(f, 'X' );
		write_int(f, count);
		for (rxa = curr_rsync_xal.rxas; count--; rxa++) {
			write_int( f, rxa->name_len );
			write_int ( f, rxa->datum_len );
			write_buf ( f, rxa->name, rxa->name_len );
			write_buf ( f, rxa->datum, rxa->datum_len );
		}
		rsync_xal_store();
	}
}


/* ------------------------------------------------------------------------- */
/* receive and build the rsync_xattr_lists */

void receive_xattr(struct file_struct *file, int f)
{
	char *fname;
	int tag;

	if ( ! preserve_xattrs )
		return;
	fname = f_name(file);
	tag = read_byte(f);
	if (tag != 'X' && tag != 'x') {
	    rprintf(FERROR, "%s: receive_xattr: unknown extended attribute type tag: %c\n",
 fname, tag);
	    exit_cleanup(RERR_STREAMIO);
	}

	if ( fxil.alloc <= fxil.count) {
		void *new_ptr;
		size_t new_alloc;
		size_t new_size;

		if ( fxil.count <  RSYNC_XAL_LIST_INITIAL )
			new_alloc = fxil.alloc + RSYNC_XAL_LIST_INITIAL;
		else
			new_alloc = fxil.alloc * 2;
		new_size = new_alloc * sizeof(file_xal_index);
		if ( fxil.alloc )
			new_ptr = realloc ( fxil.filexalidxs, new_size );
		else
			new_ptr = malloc ( new_size );
		if (!new_ptr)
			out_of_memory("receive_xattr");
		if (verbose >= 3) {
			rprintf(FINFO, "receive_xattr to %lu bytes, %s move\n",
				(unsigned long) new_size,
				fxil.filexalidxs == new_ptr ? "did not" : "did");
		}

		fxil.filexalidxs = new_ptr;
		fxil.alloc = new_alloc;
	}

	fxil.filexalidxs[fxil.count].file = file;
	if (tag == 'X' ) {
		size_t count;
		size_t i;

		fxil.filexalidxs[fxil.count].xalidx = rsync_xal_l.count;

		count = read_int ( f );
		curr_rsync_xal.count = count;
		curr_rsync_xal.alloc = count;
		curr_rsync_xal.rxas = malloc ( count * sizeof(rsync_xa) );
		if ( curr_rsync_xal.rxas == NULL ) {
			rprintf(FERROR, "%s: receive_xattr: allocate %lu xas failed\n",
 fname, (unsigned long)count );
			exit_cleanup(RERR_STREAMIO);
		}
		for ( i = 0; i < count; i++ ) {
			size_t name_len;
			size_t datum_len;
			char *ptr;

			name_len = read_int ( f );
			datum_len = read_int ( f );
			ptr = malloc ( name_len + datum_len );
			read_buf ( f, ptr, name_len );
			read_buf ( f, ptr + name_len, datum_len );
			curr_rsync_xal.rxas[i].name_len = name_len;
			curr_rsync_xal.rxas[i].datum_len = datum_len;
			curr_rsync_xal.rxas[i].name = ptr;
			curr_rsync_xal.rxas[i].datum = ptr + name_len;
		}
		rsync_xal_store();
	} else {
		size_t index;

		index = read_int(f);
		if ( index >= rsync_xal_l.count) {
			rprintf(FERROR, "%s: receive_xattr: xa index %lu out of range\n",
fname, (unsigned long)index );
			exit_cleanup(RERR_STREAMIO);
		}
		fxil.filexalidxs[fxil.count].xalidx = index;
	}
	fxil.count++;
}

static BOOL rsync_xal_set ( const char *fname, rsync_xal *x ) {
	size_t i;
	int status;
	BOOL ret;

	ret = True;
	for ( i = 0; i < x->count; i++ ) {
		status = sys_lsetxattr ( fname, x->rxas[i].name, x->rxas[i].datum, x->rxas[i].datum_len, 0 );
		if ( status < 0 ) {
			rprintf(FERROR, "%s: rsync_xal_set: lsetxattr %s failed: %s\n",
				fname, x->rxas[i].name, strerror(errno));
			ret = False;
		}
	}
	return ret;
}

/* for duplicating xattrs on backups when using backup_dir */

int dup_xattr ( const char *orig, const char *bak )
{
	int ret;

	ret = 0;
	if (!preserve_xattrs)
		return ret;

	ret = rsync_xal_get ( orig, &backup_xal );
	if ( ret == True )
		ret = rsync_xal_set ( bak, &backup_xal );
	rsync_xal_free ( &backup_xal );
	return ret;
}

void push_keep_backup_xattr ( const struct file_struct *file, const char *orig, const char *dest )
{
	if ( ! preserve_xattrs )
		return;
		
	backup_orig_file = file;
	backup_orig_fname = orig;
	backup_dest_fname = dest;
	rsync_xal_get ( orig, &backup_xal );
}

static int set_keep_backup_xal ( void ) {
	if ( ! preserve_xattrs )
		return 0;
	return rsync_xal_set ( backup_dest_fname, &backup_xal );
}

void cleanup_keep_backup_xattr ( void )
{
	if ( ! preserve_xattrs)
		return;

	backup_orig_file = NULL;
	backup_orig_fname = null_string;
	backup_dest_fname = null_string;
	rsync_xal_free ( &backup_xal );
}

static int file_xal_index_compare ( const void *x1, const void *x2 ) {
    const file_xal_index *xa1;
    const file_xal_index *xa2;

    xa1 = x1;
    xa2 = x2;
    return ( xa1->file == xa2->file) ? 0 :
 ( ( xa1->file < xa2->file ) ? -1: 1 );
}
void sort_file_xattr_index_lists ( void )
{
	if ( ! preserve_xattrs )
		return;
	qsort ( fxil.filexalidxs, fxil.count, sizeof(file_xal_index),
 file_xal_index_compare );
}


static int find_file_xal_index( const struct file_struct *file) {
	int low = 0, high = fxil.count;
	const struct file_struct *file_mid;

	if ( ! high-- ) {
		rprintf(FERROR, "find_file_xal_index: no entries\n");
		exit_cleanup(RERR_STREAMIO);
		return -1;
	}
	do {
		int mid = (high + low) / 2;
		file_mid = fxil.filexalidxs[mid].file;
		if (file_mid == file)
			return fxil.filexalidxs[mid].xalidx;
		if (file_mid > file)
			high = mid - 1;
		else
			low = mid + 1;
	} while (low < high);
	if (low == high) {
		file_mid = fxil.filexalidxs[low].file;
		if (file_mid == file)
			return fxil.filexalidxs[low].xalidx;
	}
	rprintf(FERROR,
		"find_file_xal_index: can't find entry for file in list\n");
	exit_cleanup(RERR_STREAMIO);
	return -1;
}


/* set extended attributes on rsync-ed or keep_backup-ed file */

int set_xattr ( const char *fname, const struct file_struct *file )
{
	int updated;
	int xalidx;
	rsync_xal *x;

	updated = 0;
	if ( dry_run || ! preserve_xattrs )
		return 0;
	if (file == backup_orig_file) {
		if (!strcmp(fname, backup_dest_fname))
			return set_keep_backup_xal();
	}
	xalidx = find_file_xal_index( file );
	x = &(rsync_xal_l.rxals[xalidx]);
	updated = rsync_xal_set ( fname, x );
	return updated;
}

#endif /* SUPPORT_XATTRS */
